"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

import rizin
import sys
import cle  # https://github.com/angr/cle


class RzBufferWrapper:
    def __init__(self, buf):
        self.buf = buf

    def seek(self, where):
        self.buf.seek(where, 0)

    def read(self, length=None):
        if length is None:
            cur = self.buf.seek(0, rizin.RZ_BUF_CUR)  # save pos
            end = self.buf.seek(0, rizin.RZ_BUF_END)
            length = end - cur
            self.buf.seek(cur, rizin.RZ_BUF_SET)  # restore pos
        b = bytearray(length)
        self.buf.read(b)
        return b


class CLEBinPlugin(rizin.RzBinPluginDirector):
    name = "CLESWIGPlugin"
    desc = "ELF Binary Loader using the Angr CLE library"
    author = "wingdeans"
    license = "LGPL3"

    def load_buffer(self, bf, obj, buf, sdb):
        try:
            loader = cle.Loader(RzBufferWrapper(buf))
        except Exception as e:
            print(e)
            return False
        self.loader = loader
        return True

    def baddr(self, bf):
        return self.loader.main_object.linked_base

    def symbols(self, bf):
        syms = rizin.RzList_RzBinSymbol()

        for sym in self.loader.main_object.symbols:
            binsym = rizin.RzBinSymbol()
            binsym.thisown = False
            binsym.name = sym.name
            binsym.type = rizin.RZ_BIN_TYPE_FUNC_STR
            binsym.paddr = sym.linked_addr
            binsym.vaddr = sym.rebased_addr
            binsym.size = sym.size
            syms.append(binsym)
        return syms

    def sections(self, bf):
        sections = rizin.RzList_RzBinSection()
        for section in self.loader.main_object.sections:
            binsection = rizin.RzBinSection()
            binsection.thisown = False
            binsection.name = section.name
            binsection.size = section.filesize
            binsection.vsize = section.memsize
            binsection.paddr = section.offset
            binsection.vaddr = section.vaddr
            sections.append(binsection)
        return sections

    def maps(self, bf):
        # virtual memory mappings
        maps = rizin.RzList_RzBinMap()
        for segment in self.loader.main_object.segments:
            binmap = rizin.RzBinMap()
            binmap.thisown = False
            binmap.perm = segment.flags  # must be readable
            binmap.paddr = segment.offset
            binmap.psize = segment.filesize
            binmap.vaddr = segment.vaddr
            binmap.vsize = segment.memsize
            maps.append(binmap)
        return maps

    def strings(self, bf):
        # rz_bin_file_strings
        # the last parameter enables searching for strings everywhere,
        # which we disable to limit to searching only in data sections
        return bf.strings(4, False)

    def info(self, bf):
        info = rizin.RzBinInfo()
        info.has_va = True  # has virtual addresses
        return info


plugin, plugin_struct = rizin.register_RzBinPlugin(CLEBinPlugin)

core = rizin.RzCore()
plugin_struct.thisown = False
core.bin.plugins.append(plugin_struct)

# Run rizin and load binary
core.bin.force_plugin(plugin.name)
core.file_open_load(sys.argv[1])

core.prompt_loop()



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import List, OrderedDict, Optional, overload

from clang.cindex import Cursor, CursorKind

from cparser_header import Header

enums: List["Enum"] = []
macro_enums: List["MacroEnum"] = []


class Enum:
    """
    A C enum
    """

    typedef_name: str
    fields: OrderedDict[str, str]

    def __init__(self, header: Header, *, typedef: str):
        enums.append(self)

        typedef_cursor = header.pop(CursorKind.TYPEDEF_DECL, typedef)
        self.typedef_name = typedef_cursor.spelling

        enum = typedef_cursor.underlying_typedef_type.get_declaration()
        assert (
            enum.kind == CursorKind.ENUM_DECL
        ), "Typedef underlying declaration was {struct_cursor.kind}, not ENUM_DECL"

        self.fields = OrderedDict()
        for field in enum.get_children():
            assert field.kind == CursorKind.ENUM_CONSTANT_DECL
            self.fields[field.spelling] = str(field.enum_value)


class MacroEnum:
    """
    A C enum consisting of #define's
    """

    defines: OrderedDict[str, str]

    @overload
    def __init__(self, header: Header, *defines: str):
        ...

    @overload
    def __init__(self, header: Header, *, prefix: str):
        ...

    def __init__(
        self,
        header: Header,
        *defines: str,
        prefix: Optional[str] = None,
    ):
        macro_enums.append(self)

        self.defines = OrderedDict()

        def add_definition(macro: Cursor) -> None:
            toks = [tok.spelling for tok in macro.get_tokens()]
            definition = " ".join(toks[1:])
            self.defines[macro.spelling] = definition

        if prefix:
            macro_names = []
            for name, macro in header.cursors[CursorKind.MACRO_DEFINITION].items():
                if name.startswith(prefix):
                    macro_names.append(name)
                    add_definition(macro)
            header.ignore(*macro_names)

        for define in defines:
            add_definition(header.pop(CursorKind.MACRO_DEFINITION, define))



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

# This is an example designed to show the internal steps taken
# by the helper functions. To see the recommended way of doing
# things, see `4b-rz_cmd.py`

import rizin

core = rizin.RzCore()

class CustomCommand(rizin.CmdDirector):
    def run(self, core, argc, argv):
        print("hello from python")
        return True

# Group desc help
user_group_desc_help = rizin.RzCmdDescHelp()
user_group_desc_help.thisown = False
user_group_desc_help.summary = "A custom group for user-defined commands"
core.rcmd.register_swig_command("u", None, None, user_group_desc_help)

# Command desc help
custom_desc_help = rizin.RzCmdDescHelp()
custom_desc_help.thisown = False

custom_args = rizin.Array_RzCmdDescArg(1)
custom_args.thisown = False
null_arg = rizin.RzCmdDescArg()
null_arg.thisown = False
custom_args[0] = null_arg

custom_desc_help.args = custom_args.cast()

# Construct and set
custom_cmd = CustomCommand()
custom_cmd.__disown__()
core.rcmd.register_swig_command("uh", custom_cmd, custom_desc_help)

core.prompt_loop()



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import List, Dict, Set, Optional, overload, TYPE_CHECKING

from binding_typemap import Typemap
from binding_generic_specializations import gen_ctype_specializations

if TYPE_CHECKING:
    from cparser_header import Header, CFunc


class Func:
    """
    A wrapped C function
    """

    cfunc: "CFunc"
    typemaps: List[Typemap]

    @overload
    def __init__(
        self,
        header: "Header",
        name: str,
        *,
        default_args: Optional[Dict[str, str]] = ...,
        typemaps: Optional[List[Typemap]] = ...,
    ):
        ...

    @overload
    def __init__(
        self,
        header: "Header",
        *,
        cfunc: "CFunc",
        default_args: Optional[Dict[str, str]] = ...,
        typemaps: Optional[List[Typemap]] = ...,
    ):
        ...

    def __init__(
        self,
        header: "Header",
        name: Optional[str] = None,
        *,
        cfunc: Optional["CFunc"] = None,
        default_args: Optional[Dict[str, str]] = None,
        typemaps: Optional[List[Typemap]] = None,
    ):
        if name:
            cfunc = header.pop_func(name)
        assert cfunc

        for arg in cfunc.args:
            arg_name = arg.cursor.spelling
            if default_args and arg_name in default_args:
                arg.default = default_args[arg_name]
                default_args.pop(arg_name)
        assert not default_args or len(default_args) == 0

        self.cfunc = cfunc
        self.typemaps = typemaps or []
        self.gen_ctype_specializations()

        for typemap in self.typemaps:
            typemap.check(self)

    def gen_ctype_specializations(self) -> None:
        """
        Generate generic specializations for args and return type
        """
        for arg in self.cfunc.args:
            gen_ctype_specializations([arg.cursor], arg.ctype)
        gen_ctype_specializations([self.cfunc.cursor], self.cfunc.result_ctype)


class GenericFunc(Func):
    """
    A wrapped C function with generic args or a generic return value
    """

    generic_ret: bool
    generic_args: Set[str]

    def __init__(
        self,
        header: "Header",
        name: str,
        *,
        generic_ret: bool = False,
        generic_args: Optional[Set[str]] = None,
    ):
        self.generic_ret = generic_ret
        self.generic_args = generic_args or set()
        super().__init__(header, name)

    def gen_ctype_specializations(self) -> None:
        if not self.generic_ret:
            gen_ctype_specializations([self.cfunc.cursor], self.cfunc.result_ctype)

        generic_args = self.generic_args.copy()

        for arg in self.cfunc.args[1:]:
            arg_name = arg.cursor.spelling
            if arg_name in generic_args:
                generic_args.remove(arg_name)
            else:
                gen_ctype_specializations([arg.cursor], arg.ctype)

        # Ensure all generic_args are for valid args
        assert len(generic_args) == 0



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import Iterator, TextIO

import os
from contextlib import contextmanager


class Writer:
    """
    Helper class for writing indented lines to an output
    """

    output: TextIO
    indent_level: int
    indent_amount: int

    def __init__(self, output: TextIO, *, indent_amount: int = 4):
        self.output = output
        self.indent_level = 0
        self.indent_amount = indent_amount

    def line(self, *lines: str) -> None:
        """
        Write lines at current indentation
        """
        for line in lines:
            indent = self.indent_amount * self.indent_level
            self.output.write(" " * indent)
            self.output.write(line)
            self.output.write("\n")

    @contextmanager
    def indent(self) -> Iterator[None]:
        """
        Increase indentation level for duration of context
        """
        self.indent_level += 1
        yield
        self.indent_level -= 1

    def snippet(self, path: str) -> None:
        """
        Write a snippet file at current indentation

        File searched from this file's directory
        """
        path_segments = path.split("/")
        filename = os.path.join(os.path.dirname(__file__), *path_segments)
        with open(filename, encoding="utf-8") as snippet:
            self.line(*snippet.read().splitlines())



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import Optional, cast

import os
import shlex
from argparse import ArgumentParser

import cparser_header
import bindings

from clang.cindex import Config

parser = ArgumentParser()
parser.add_argument("--output-dir", "-o", required=True)
parser.add_argument("--clang-path", required=True)
parser.add_argument("--clang-args", required=True)
parser.add_argument("--rizin-include-path", required=True)
parser.add_argument("--targets", required=True)
parser.add_argument("--doxygen-path")
args = parser.parse_args()

output_dir = cast(str, args.output_dir)
Config.set_library_path(cast(str, args.clang_path))
cparser_header.clang_args = clang_args = shlex.split(cast(str, args.clang_args))
cparser_header.rizin_include_path = rizin_include_path = cast(
    str, args.rizin_include_path
)
targets = set(cast(str, args.targets).split(","))

# Add additional include directories
for segments in [
    # already isntalled `include/librz` directory
    ["."],
    ["sdb"],
    # not yet installed meson project `include` directory
    ["..", "..", "build"],
    ["..", "util", "sdb", "src"],
]:
    path = os.path.abspath(os.path.join(rizin_include_path, *segments))
    if os.path.exists(path):
        clang_args += ["-I", path]

# Enable certain annotation definitions in rz_types.h
clang_args.append("-DRZ_BINDINGS")

# Run binding specifications
bindings.run()

# Generator(s)
if "SWIG" in targets:
    import generator_swig

    generator_swig.generate(output_dir)

if "sphinx" in targets:
    import generator_sphinx

    generator_sphinx.doxygen_path = cast(Optional[str], args.doxygen_path)
    generator_sphinx.generate(output_dir)



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import List, Dict, OrderedDict

from dataclasses import dataclass

from cparser_header import Header
from cparser_types import CType, CPointerType, CFunctionType
from binding_class import Class

directors: Dict[str, "Director"] = {}


@dataclass
class DirectorArg:
    """
    Groups a parameter cursor and its type
    """

    name: str
    ctype: CType


@dataclass
class DirectorFunc:
    """
    Groups a function cursor's args and return types
    """

    args: List[DirectorArg]
    result_ctype: CType


class Director:
    """
    A struct with function pointers that can be set from the guest language
    """

    name: str

    fields: OrderedDict[str, CType]
    funcs: OrderedDict[str, DirectorFunc]

    def __init__(self, header: Header, typedef: str):
        cls = Class(header, typedef=typedef)

        self.name = typedef
        self.fields = OrderedDict()
        self.funcs = OrderedDict()

        assert typedef not in directors
        directors[typedef] = self

        for name, field in cls.fields.items():
            if isinstance(field.ctype, CPointerType) and isinstance(
                field.ctype.pointee, CFunctionType
            ):
                ctype = field.ctype.pointee

                # arg_names should be generated in Class construction
                # when calling gen_ctype_specializations on field
                assert ctype.arg_names and len(ctype.arg_names) == len(ctype.args)

                self.funcs[name] = DirectorFunc(
                    [
                        DirectorArg(arg_name, arg_ctype)
                        for arg_name, arg_ctype in zip(ctype.arg_names, ctype.args)
                    ],
                    ctype.result,
                )
            else:
                self.fields[name] = field.ctype



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only

This is a Rizin linter script, intended for use in CI.
It parses the compile_commands.json created by meson,
then runs libclang on those files with the specified arguments.

It detects TYPEREF cursors in function parameters, function return types
and struct fields. If the TYPEREF spelling is one of the prespecified
generic types, expect a /*<type>*/ comment and emits a warning if none is found.

Certain generic types, such as RzList, must include a pointer in their comment.
This is since many RzList specializations are for char*, pointing to a string's starting
character, and eliding the pointer would result in an RzList /*<char>*/, which loses
the implied meaning of holding a string.

Other generic types, such as RzVector, should not contain a pointer (RzPVector
should be used in cases where something must be pointed to).

The linter also enforces consistency in type comments between function declarations
and definitions. It also does this for Rizin annotations, such as RZ_OWN. This works
by defining the RZ_BINDINGS preprocessor flag, which sets the annotations to expand to
__attribute__((annotate)), which can be picked up by libclang.
"""
from typing import List, Dict, Set, TypedDict, Optional, cast

import os
import sys
import json
import shlex
from argparse import ArgumentParser
from dataclasses import dataclass
from itertools import zip_longest

from clang.cindex import (
    Config,
    TranslationUnit,
    TranslationUnitLoadError,
    Cursor,
    CursorKind,
    SourceRange,
    SourceLocation,
)


warnings = set()


def warn(warning: str) -> None:
    """
    Print a warning only once
    """
    if warning not in warnings:
        print(warning)
        warnings.add(warning)


def stringify_location(location: SourceLocation) -> str:
    """
    Get <relpath:filename:line> for a SourceLocation
    """
    path = os.path.relpath(os.path.abspath(location.file.name))
    return f"<{path}:{location.line}:{location.column}>"


def cursor_get_annotations(cursor: Cursor) -> List[str]:
    """
    Get RZ_* annotation attributes on a cursor
    """
    return [
        child.spelling
        for child in cursor.get_children()
        if child.kind == CursorKind.ANNOTATE_ATTR and child.location in cursor.extent
    ]


generic_types = {"RzList", "RzListIter", "RzPVector", "RzVector", "RzGraph"}


def cursor_get_comment(cursor: Cursor, *, packed: bool = False) -> Optional[str]:
    """
    Get /*<type>*/ comment on a cursor

    The packed argument is used to mark a struct field is marked by
    the RZ_PACK macro, so that the /*<type>*/ comment is searched for
    at the original location within the macro call
    """

    assert cursor.kind in [
        CursorKind.FUNCTION_DECL,
        CursorKind.PARM_DECL,
        CursorKind.FIELD_DECL,
    ]

    try:
        typeref = next(
            child
            for child in cursor.get_children()
            if child.kind == CursorKind.TYPE_REF
        )
        typeref_spelling = typeref.spelling
    except StopIteration:
        return None

    if packed:
        assert cursor.kind == CursorKind.FIELD_DECL
        src_range = SourceRange.from_locations(typeref.extent.end, cursor.location)
    else:
        # Reconstruct SourceLocation without macro bit
        start, end, translation_unit = (
            typeref.extent.end,
            cursor.location,
            cursor.translation_unit,
        )
        start = SourceLocation.from_position(
            translation_unit, start.file, start.line, start.column
        )
        end = SourceLocation.from_position(
            translation_unit, end.file, end.line, end.column
        )
        src_range = SourceRange.from_locations(start, end)

    # Ensure comment token exists
    try:
        token = next(cursor.translation_unit.get_tokens(extent=src_range))
    except StopIteration:
        if typeref_spelling in generic_types:
            warn(
                f"Missing type comment at {stringify_location(cursor.location)} "
                f"for {typeref_spelling}"
            )
        return None

    # Filter out other tokens
    comment = token.spelling
    if not comment.startswith("/*") or not comment.endswith("*/"):
        if typeref_spelling in generic_types:
            warn(
                f"Missing type comment at {stringify_location(cursor.location)} "
                "(token is not a comment)"
            )
        return None
    comment = comment[2:-2]

    if not comment.startswith("<") or not comment.endswith(">"):
        if typeref_spelling in generic_types:
            warn(
                f"Type comment at {stringify_location(cursor.location)} lacks angle brackets"
            )
        return comment

    # Check pointer (or lack of) and space between pointer
    if typeref_spelling in {"RzList", "RzListIter", "RzPVector", "RzGraph"}:
        if comment[-2] != "*":
            warn(f"Type comment at {stringify_location(cursor.location)} lacks pointer")
        elif comment[-3] != " ":
            warn(
                f"Type comment at {stringify_location(cursor.location)} lacks space between pointer"
            )
    elif typeref_spelling in {"RzVector"}:
        if comment[-2] == "*":
            warn(
                f"Type comment at {stringify_location(cursor.location)} should not have pointer"
            )
    elif typeref_spelling in {"HtPP", "HtUP", "HtUU", "RBTree", "SdbList"}:
        pass
    else:
        warn(
            f"Type comment at {stringify_location(cursor.location)} "
            f"for unknown type {typeref_spelling}"
        )

    return comment


@dataclass
class Arg:
    """
    Groups an argument's annotations and comment
    """

    annotations: List[str]
    comment: Optional[str]


class Function:
    """
    Represents a libclang function cursor and facilitates diffing with other functions
    """

    name: str
    annotations: List[str]
    comment: Optional[str]
    args: List[Arg]
    location: str

    def __init__(self, cursor: Cursor):
        self.name = cursor.spelling
        self.location = stringify_location(cursor.location)

        self.annotations = cursor_get_annotations(cursor)
        self.comment = cursor_get_comment(cursor)

        self.args = [
            Arg(
                annotations=cursor_get_annotations(arg), comment=cursor_get_comment(arg)
            )
            for arg in cursor.get_arguments()
        ]

    def diff(self, new: "Function") -> None:
        """
        Compare annotations and comments on function return type and arguments.

        Warn if mismatch
        """

        for annotation, new_annotation in zip_longest(
            self.annotations, new.annotations, fillvalue=None
        ):
            if annotation != new_annotation:
                warn(
                    f"Mismatched function annotation for {new.name} "
                    f"at {new.location} : {new_annotation} "
                    f"(was {annotation} at {self.location}"
                )

        if self.comment != new.comment:
            warn(
                f"Mismatched function type comment for {new.name} "
                f"at {new.location} : {new.comment} "
                f"(was {self.comment} at {self.location})"
            )

        for arg, new_arg in zip_longest(self.args, new.args, fillvalue=None):
            assert arg and new_arg

            for annotation, new_annotation in zip_longest(
                arg.annotations, new_arg.annotations, fillvalue=None
            ):
                if annotation != new_annotation:
                    warn(
                        f"Mismatched function annotation for {new.name} "
                        f"at {new.location} : {new_annotation} "
                        f"(was {annotation} at {self.location}"
                    )

                if arg.comment != new_arg.comment:
                    warn(
                        f"Mismatched function argument type comment for {new.name} "
                        f"at {new.location} : {new.comment} "
                        f"(was {self.comment} at {self.location})"
                    )


def check_translation_unit(
    translation_unit: TranslationUnit, *, skipped_paths: Set[str], rizin_path: str
) -> None:
    """
    Check for issues in translation_unit
    """

    for diagnostic in translation_unit.diagnostics:
        warn(
            f"Translation unit diagnostic at "
            f"{stringify_location(diagnostic.location)}: {diagnostic.spelling}"
        )

    functions: Dict[str, Function] = {}

    for cursor in translation_unit.cursor.get_children():
        cursor_file = cursor.location.file
        if not cursor_file:
            continue

        abspath = os.path.abspath(cursor_file.name)
        if not abspath.startswith(rizin_path):
            continue

        if abspath in skipped_paths:
            continue

        if cursor.kind == CursorKind.FUNCTION_DECL:
            function = Function(cursor)

            if function.name in functions:
                function.diff(functions[function.name])
            functions[function.name] = function
        elif cursor.kind == CursorKind.STRUCT_DECL:
            packed = False
            for field in cursor.get_children():
                if field.kind == CursorKind.PACKED_ATTR:
                    packed = True
                elif field.kind == CursorKind.FIELD_DECL:
                    cursor_get_comment(field, packed=packed)
                elif field.kind not in [
                    CursorKind.STRUCT_DECL,
                    CursorKind.UNION_DECL,
                    CursorKind.ENUM_DECL,
                ]:
                    warn(f"Unknown field cursor kind: {field.kind}")


class Command(TypedDict):
    """
    A single entry in compile_commands.json
    """

    file: str
    directory: str
    command: str


def main() -> int:
    """
    CLI Entrypoint
    """
    parser = ArgumentParser()
    parser.add_argument("--clang-path", required=True)
    parser.add_argument("--clang-args", required=True)
    parser.add_argument("--rizin-path", required=True)
    args = parser.parse_args()

    Config.set_library_path(cast(str, args.clang_path))
    clang_base_args = shlex.split(cast(str, args.clang_args)) + ["-DRZ_BINDINGS"]
    rizin_path = os.path.abspath(cast(str, args.rizin_path))

    skipped_paths = {
        os.path.join(rizin_path, "librz", *segments)
        for segments in [
            ["include", "rz_list.h"],
            ["util", "list.c"],
            ["include", "rz_skiplist.h"],
            ["util", "skiplist.c"],
            ["include", "rz_vector.h"],
            ["util", "vector.c"],
            ["include", "rz_util", "rz_graph.h"],
            ["util", "graph.c"],
            ["include", "rz_util", "rz_idpool.h"],
            ["util", "idpool.c"],
            ["util", "regex", "regcomp.c"],
            ["include", "rz_magic.h"],
        ]
    }

    # Used to parse compile_commands command entry
    cmd_parser = ArgumentParser()
    cmd_parser.add_argument("-I", action="append")
    cmd_parser.add_argument("-D", action="append")

    with open(
        os.path.join(rizin_path, "build", "compile_commands.json"), encoding="utf-8"
    ) as compile_commands:
        commands: List[Command] = json.loads(compile_commands.read())

        for command in commands:
            abspath = os.path.abspath(
                os.path.join(command["directory"], command["file"])
            )
            relpath = os.path.relpath(abspath, rizin_path)

            if relpath.startswith("subproject") or relpath.startswith("test"):
                continue

            namespace, _ = cmd_parser.parse_known_args(shlex.split(command["command"]))
            defines = cast(List[str], namespace.D)
            includes = cast(List[str], namespace.I)

            clang_args = clang_base_args.copy()
            clang_args += ["-D" + define for define in defines]
            clang_args += [
                "-I" + os.path.join(command["directory"], include)
                for include in includes
            ]

            try:
                check_translation_unit(
                    TranslationUnit.from_source(abspath, clang_args),
                    skipped_paths=skipped_paths,
                    rizin_path=rizin_path,
                )
            except TranslationUnitLoadError:
                warn(f"Failed to parse file {relpath}")

    return len(warnings)


if __name__ == "__main__":
    sys.exit(main())



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

# This is an example designed to show the internal steps taken
# by the helper functions. To see the recommended way of doing
# things, see `2b-rz_bin_plugin.py`

import rizin
import sys


import rizin

class CustomBinPlugin(rizin.RzBinPluginDirector):
    def __init__(self):
        super().__init__()
        self.__disown__()

    def load_buffer(self, bf, obj, buf, sdb):
        print("name: ", bf.file)
        print("size: ", bf.size)
        return True

# Construct the director
builder = rizin.RzBinPluginBuilder()
builder.name = "SWIGCustom"
builder.enable_load_buffer = True
plugin = builder.build(CustomBinPlugin())

core = rizin.RzCore()
core.bin.plugins.append(plugin)

# Run rizin and load binary
core.bin.force_plugin(plugin.name)
core.file_open_load(sys.argv[1])

while True:
    core.flush(input("> "))



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""
from typing import List, Dict, TYPE_CHECKING

from clang.cindex import Cursor, CursorKind

from cparser_types import (
    CType,
    CPointerType,
    CRecordType,
    CFunctionType,
    CArrayType,
    CTypedefType,
    CPrimitiveType,
    assert_never,
)

if TYPE_CHECKING:
    from binding_generic import Generic

generic_structs: Dict[str, "Generic"] = {}


def gen_ctype_specializations(cursors: List[Cursor], ctype: CType) -> None:
    """
    Generates necessary specializations for a cursor and type

    Multiple cursors are provided as input to accomodate for typedefs
    which have their own type comments
    """
    if isinstance(ctype, CPointerType):
        gen_ctype_specializations(cursors, ctype.pointee)
    elif isinstance(ctype, CTypedefType):
        if isinstance(ctype.canonical, CRecordType):
            gen_ctype_specializations(cursors, ctype.canonical)
        else:
            # Add typedef cursor for non-struct typedefs
            # so type comments can be processed
            # eg. typedef RzVector /*<ut64>*/ (*func)(void)
            gen_ctype_specializations(cursors + [ctype.cursor], ctype.canonical)
    elif isinstance(ctype, CFunctionType):
        gen_ctype_specializations(cursors, ctype.result)
        cursor_args = [
            cursor
            for cursor in cursors[-1].get_children()
            if cursor.kind == CursorKind.PARM_DECL
        ]

        assert len(cursor_args) == len(ctype.args)
        arg_names = []
        for arg_cursor, arg_ctype in zip(cursor_args, ctype.args):
            gen_ctype_specializations([arg_cursor], arg_ctype)
            arg_names.append(arg_cursor.spelling)
        ctype.arg_names = arg_names
    elif isinstance(ctype, CRecordType):
        decl_spelling = ctype.decl_spelling
        if decl_spelling in generic_structs:
            ctype.generic = generic_structs[decl_spelling]

            # Search for comment in all active cursors
            for cursor in cursors:
                ctype.specialization = ctype.generic.add_specialization(cursor)
                if ctype.specialization:
                    break
            else:
                raise Exception(
                    "No /*<type>*/ comment found in cursors at "
                    + ", ".join(str(cursor.location) for cursor in cursors)
                )
    elif isinstance(ctype, CArrayType):
        pass
    elif isinstance(ctype, CPrimitiveType):
        pass
    else:
        assert_never(ctype)



def register_group(self, cmd, summary):
    help_desc = RzCmdDescHelp()
    help_desc.thisown = False
    help_desc.summary = summary
    self.rcmd.register_swig_command(cmd, None, None, help_desc)

def register_command(self, cmd, fn):
    import inspect
    params = list(inspect.signature(fn).parameters.values())

    core_arg = None
    if len(params) > 0 and params[0].annotation == RzCore:
        core_arg = params[0].name
        params = params[1:]

    desc_args = Array_RzCmdDescArg(len(params) + 1)
    desc_args.thisown = False
    for i, param in enumerate(params):
        if not param.annotation:
            raise Exception(f"Parameter {param.name} has no annotation")

        desc_arg = RzCmdDescArg()
        desc_arg.name = param.name

        if param.annotation == str:
            desc_arg.type = RZ_CMD_ARG_TYPE_STRING
        elif param.annotation == RzNumArg:
            desc_arg.type = RZ_CMD_ARG_TYPE_RZNUM
        elif param.annotation == int:
            desc_arg.type = RZ_CMD_ARG_TYPE_NUM
        elif param.annotation == RzFilenameArg:
            desc_arg.type = RZ_CMD_ARG_TYPE_FILE
        elif param.annotation == RzFlagItem:
            desc_arg.type = RZ_CMD_ARG_TYPE_FLAG
        elif param.annotation == RzAnalysisFunction:
            desc_arg.type = RZ_CMD_ARG_TYPE_FCN
        else:
            raise Exception(
                f"Parameter {param.name} has unknown type {param.annotation}"
            )

        desc_arg.thisown = False
        desc_args[i] = desc_arg

    null_arg = RzCmdDescArg()
    null_arg.thisown = False
    desc_args[len(params)] = null_arg

    help_desc = RzCmdDescHelp()
    help_desc.thisown = False
    help_desc.args = desc_args.cast()

    class wrapper(CmdDirector):
        def run(self, core, argc, argv):
            try:
                args = {}
                args_array = Array_String.frompointer(argv)
                if core_arg:
                    args[core_arg] = core
                for i, param in enumerate(params):
                    orig_arg = args_array[i + 1]
                    arg = orig_arg
                    if param.annotation == RzNumArg:
                        if not core.num.is_valid_input(arg):
                            raise ValueError(f"Could not convert '{orig_arg}' to RzNum value")

                        arg = core.num.math(arg)
                    elif param.annotation == int:
                        try:
                            arg = int(arg)
                        except ValueError:
                            raise ValueError(f"Could not convert '{orig_arg}' to int") from e
                    elif param.annotation == RzFlagItem:
                        arg = core.flags.get(arg)
                        if arg is None:
                            raise ValueError(f"Could not find a flag named '{orig_arg}'")
                    elif param.annotation == RzAnalysisFunction:
                        arg = core.analysis.get_function_byname(arg)
                        if arg is None:
                            raise ValueError(f"Could not find a function named '{orig_arg}'")
                    args[param.name] = arg
                return fn(**args)
            except Exception as e:
                print(e)
                return False

    director = wrapper()
    director.__disown__()
    self.rcmd.register_swig_command(cmd, director, help_desc)



class RzListIterator:
    def __init__(self, rzlist):
        self.iter = rzlist.iterator()

    def __next__(self):
        if self.iter is None:
            raise StopIteration
        data = self.iter.data()
        self.iter = self.iter.next()
        return data


class RzVectorIterator:
    def __init__(self, rzvector):
        self.rzvector = rzvector
        self.index = 0

    def __next__(self):
        if self.index >= len(self.rzvector):
            raise StopIteration
        data = self.rzvector.index_ptr(self.index)
        self.index += 1
        return data


class RzPVectorIterator:
    def __init__(self, rzpvector):
        self.rzpvector = rzpvector
        self.index = 0

    def __next__(self):
        if self.index >= len(self.rzpvector):
            raise StopIteration
        data = self.rzpvector.at(self.index)
        self.index += 1
        return data
